@@ -1,4 +1,8 @@
 name = 'ldf_adapter'
+# vim: foldmethod=indent : tw=100
+# pylint: disable=invalid-name, superfluous-parens
+# pylint: disable=logging-fstring-interpolation, logging-not-lazy, logging-format-interpolation
+# pylint: disable=missing-docstring, too-few-public-methods
 
 import logging
 from collections import Mapping
@@ -14,7 +18,8 @@ from . import eduperson
 
 from . import backend
 from .config import CONFIG
-from .results import Deployed, NotDeployed, Rejection, Failure, Question, raise_question
+from .results import Deployed, NotDeployed, Rejection, Failure, Question, raise_question, Status
+from .name_generators import FriendlyNameGenerator
 
 logger = logging.getLogger(__name__)
 
@@ -28,6 +33,7 @@ class User:
     __init__ for details).
     """
     def __init__(self, data):
+        logger.info("init")
         """
         Arguments:
         data -- Information about the user (type: UserInfo or dict)
@@ -41,18 +47,12 @@ class User:
         self.service_user = backend.User(self.data)
         self.service_groups = [backend.Group(grp) for grp in self.data.groups]
 
+        for g in self.service_groups:
+            logger.debug(F"    group: {g.name}")
+
         if self.service_user.exists():
             self.update_username_from_existing()
-
-        logger.info("/--------------------------------------------------------------------------------\\")
-        logger.info(F"primary group: {UserInfo(data).primary_group}")
-        logger.info(F"primary group from service_user: {self.service_user.primary_group}")
-        logger.info("--------------------------------------------------------------------------------")
-
-        logger.info("self.service_groups: ")
-        for group in self.service_groups:
-            logger.info(F"    {group.name}")
-        logger.info("\\--------------------------------------------------------------------------------/")
+        logger.info("init done")
 
     def assurance_verifier(self):
         """Produce a suitably function to check if a user is allowed.
@@ -73,6 +73,7 @@ class User:
         """
         ass = CONFIG['assurance']
         prefix = ass['prefix']
+        prefix=prefix.rstrip('/') + '/'
 
         tokens = regex.findall('&|\||\(|\)|[^\s()&|]+', ass['require'])
 
@@ -155,17 +156,28 @@ class User:
         target -- The desired state. One of 'deployed' and 'not_deployed'.
         user -- The user to be deployed/undeployed (type: User)
         """
-        if not CONFIG.get('assurance', 'skip', fallback="No") =="Yes, do as I say!":
-            if not self.assurance_verifier()(self.data.assurance):
-                if not CONFIG.getboolean('assurance', 'verified_undeploy', fallback=False) and target == 'not_deployed':
-                    logger.warning("Assurance level is insufficient. Undeploying anyway.")
-                else:
-                    raise Rejection(message="Your assurance level is insufficient to access this resource")
+        try:
+            logger.info(F"Incoming request to '{target}' user '{self.data.email}' ({self.service_user.unique_id})")
+        except AttributeError:
+            logger.info(F"Incoming request to '{target}' user '{self.data.full_name}' ({self.service_user.unique_id})")
+        except AttributeError:
+            logger.info(F"Incoming request to '{target}' user '{self.service_user.unique_id}'")
 
         if target == 'deployed':
+            if not CONFIG.get('assurance', 'skip', fallback="No") =="Yes, do as I say!":
+                if not self.assurance_verifier()(self.data.assurance):
+                    raise Rejection(message="Your assurance level is insufficient to access this resource")
             return self.deploy()
         elif target == 'not_deployed':
+            if not CONFIG.get('assurance', 'skip', fallback="No") =="Yes, do as I say!":
+                if not self.assurance_verifier()(self.data.assurance):
+                    if not CONFIG.getboolean('assurance', 'verified_undeploy', fallback=False):
+                        logger.warning("Assurance level is insufficient. Undeploying anyway.")
+                    else:
+                        raise Rejection(message="Your assurance level is insufficient to access this resource")
             return self.undeploy()
+        elif target == 'get_status':
+            return self.get_status()
         else:
             raise ValueError(f"Invalid target state: {target}")
 
@@ -186,6 +198,9 @@ class User:
         if was_created:
             what_changed += 'User was created'
         else:
+            # FIXME: a user that was not created might not exist for other reasons.
+            #        the code probably relies on Failures rosen.
+            #        A "pending" flow might require additional Classes for return
             what_changed += 'User already existed'
 
         if new_groups:
@@ -207,12 +222,59 @@ class User:
 
         what_changed = ''
         if was_removed:
-            what_changed += 'User was removed.'
+            what_changed += F"User '{self.service_user.name}' was removed."
         else:
-            what_changed += 'User didn\'t exist.'
+            what_changed += F"No user for '{self.service_user.unique_id}' existed. "+\
+                            F"User '{self.service_user.name}' was not changed"
 
         return NotDeployed(message=what_changed)
 
+    def get_status(self):
+        """
+        Return the current status (that he has in the underlying local user management system)
+        User can have these status:
+        +--------------+-----------------------------------------------------------------+-----------------+
+        | Status       | Comment                                                         | Backend support |
+        +--------------+-----------------------------------------------------------------+-----------------+
+        +--------------+-----------------------------------------------------------------+-----------------+
+        | deployed     | There is an account for the user identified by unique_id        | Mandatory       |
+        +--------------+-----------------------------------------------------------------+-----------------+
+        | not deployed | There is no account for the user identified by unique_id        | Mandatory       |
+        |              | We have no information if there has ever been an account        |                 |
+        +--------------+-----------------------------------------------------------------+-----------------+
+        | rejected     | This might not be supportable; Depends on the backend           | Optional        |
+        +--------------+-----------------------------------------------------------------+-----------------+
+        | suspended    | The user with unique_id has been suspended                      | Optional        |
+        +--------------+-----------------------------------------------------------------+-----------------+
+        | pending      | The creation of the user is pending                             | Optional        |
+        +--------------+-----------------------------------------------------------------+-----------------+
+        | expired      | The user was expired, typically after being idle for some time  | Optional        |
+        +--------------+-----------------------------------------------------------------+-----------------+
+        | unknown      | We don't know the status, but at least the user is not deployed | Mandatory       |
+        +--------------+-----------------------------------------------------------------+-----------------+
+        """
+        
+        msg="No message"
+        if self.service_user.exists():
+            msg=F"username {self.service_user.name}"
+            return Status("deployed", message=msg)
+        if hasattr(self.service_user, "is_rejected"):
+            if self.service_user.is_rejected():
+                return Status("rejected", message=msg)
+        if hasattr(self.service_user, "is_suspended"):
+            if self.service_user.is_suspended():
+                return Status("suspended", message=msg)
+        if hasattr(self.service_user, "is_pending"):
+            if self.service_user.is_pending():
+                return Status("pending", message=msg)
+        if hasattr(self.service_user, "is_expired"):
+            if self.service_user.is_expired():
+                return Status("expired", message=msg)
+        if not self.service_user.exists():
+            return Status("not_deployed", message=msg)
+        return Status("unknown", message=msg)
+        
+
     def ensure_exists(self):
         """Ensure that the user exists on the service.
 
@@ -229,21 +291,33 @@ class User:
         is_new_user = not self.service_user.exists()
 
         if is_new_user:
-            logger.info('Creating user {username} for {unique_id}'.format(**self.data))
+            username = self.service_user.name
+            logger.info(F'self.service_user.name: {self.service_user.name}')
+            unique_id= self.service_user.unique_id
+            logger.info(F'Creating user "{username}" for "{unique_id}"')
 
-            if self.service_user.name_taken():
-                logger.info(F'Username {username} is already taken, asking user to pick a new one')
-                raise Question(
-                    name='username',
-                    text='Username {} already taken on this service. Please enter another one.'.format(
-                        self.data.username
+            # Raise question in case of existing username in case we're interactive
+            if CONFIG.getboolean('ldf_adapter', 'interactive', fallback=False):
+                if self.service_user.name_taken():
+                    logger.info(F'Username "{username}" is already taken, asking user to pick a new one')
+                    raise Question(
+                        name='username',
+                        text=F'Username "{username}" already taken on this service. Please enter another one.'
                     )
-                )
-            self.service_user.create()
+            else:
+                logger.info("starting while")
+                fng = FriendlyNameGenerator(self.data)
+                logger.info(F"initial try: {self.service_user.name}")
+                while self.service_user.name_taken():
+                    self.service_user.name = fng.suggest_name(self.service_user.name)
+                logger.info(F'                             Using: {self.service_user.name}')
+                if self.service_user.name is None:
+                    raise Rejection(message=F"I cannot create usernames. "
+                                    F"The list of tried ones is: {', '.join(fng.tried_names())}.")
+                self.service_user.create()
         else: # The user exists
             # Update service_user.name if unique_id already points to a username:
-            logger.debug('User for {unique_id} already exists. Nothing to do.'.format(**self.data))
-
+            logger.debug('User for "{unique_id}" already exists. Nothing to do.'.format(**self.data))
 
         self.service_user.update()
         return is_new_user
@@ -269,12 +343,14 @@ class User:
         Return True, if the user existed before.
         """
         if self.service_user.exists():
-            logger.info('Deleting user {username} of {unique_id}'.format(**self.data))
+            self.service_user.username = self.service_user.get_username()
+            logger.info(F"Deleting user '{self.service_user.username}' ({self.data.unique_id})")
+            self.service_user.username = self.service_user.get_username()
             self.service_user.uninstall_ssh_keys()
             self.service_user.delete()
             return True
         else:
-            logger.debug('No user for {unique_id} did exist. Nothing to do.'.format(**self.data))
+            logger.debug(F'No user for {self.data.unique_id} did exist. Nothing to do.')
             return False
 
     def ensure_groups_exist(self):
@@ -351,10 +427,11 @@ class UserInfo(Mapping):
         self.userinfo = data['user']['userinfo']
         self.answers = data.get('answers', {})
         self.credentials = data['user'].get('credentials', {})
+        self.allow_question = CONFIG.getboolean('ldf_adapter', 'interactive', fallback=False)
 
     @property
     @lru_cache(maxsize=None)
-    def unique_id(self, allow_question=True):
+    def unique_id(self):
         """Globally and uniquely identifies the user.
 
         This can be easily used to find out the identity of the user in the data source.
@@ -375,7 +452,7 @@ class UserInfo(Mapping):
 
     @property
     @lru_cache(maxsize=None)
-    def eppn(self, allow_question=True):
+    def eppn(self):
         """Uniquely identifies the user.
 
         At least almost. Due to homogenisations, there might be collisions. E.g. the following users
@@ -403,8 +480,9 @@ class UserInfo(Mapping):
         sub = regex.sub('[^a-zA-Z0-9_!#$%&*+/=?{|}~^.-]', '-', sub)
 
         if sub != self.userinfo['sub']:
-            logger.warning("Subject '{}' changed to '{}' for BWIDM compatibilty".format(
-                self.userinfo['sub'], sub))
+            if CONFIG.getboolean('messages', 'log_name_changes', fallback=True):
+                logger.warning("Subject '{}' changed to '{}' for BWIDM compatibilty".format(
+                    self.userinfo['sub'], sub))
 
         return sub
 
@@ -422,37 +500,46 @@ class UserInfo(Mapping):
         # We don't consider stripping the http[s]-prefix a change, since we always do that anyway,
         # and there shouldn't be two different issuers `http://example.org' and `https://example.org'.
         if iss != stripped_iss:
-            logger.warning("Issuer '{}' changed to '{}' for BWIDM compatibilty".format(
-                stripped_iss, iss))
+            if CONFIG.getboolean('messages', 'log_name_changes', fallback=True):
+                logger.warning("Issuer '{}' changed to '{}' for BWIDM compatibilty".format(
+                    stripped_iss, iss))
 
         return iss
 
     @property
     @lru_cache(maxsize=None)
-    def username(self, allow_question=True):
+    def username(self):
         """Return the user's preferred username, or ask for one if none was provided."""
-        return self.value_or_ask(
-            self.userinfo.get('preferred_username'), 'username',
-            'You have not set a global username preference. Please enter your preferred username.',
-            allow_question
-        )
+        # FIXME: This function is called, even when there is already a local user with an existing
+        # name.  SImply not having a "preferred_username" does not mean that we have to bother the user!!!
+        if self.allow_question:
+            return self.value_or_ask(
+                self.userinfo.get('preferred_username'), 'username',
+                'You have not set a global username preference. Please enter your preferred username.',
+                self.allow_question
+            )
+        fng = FriendlyNameGenerator(self)
+        return fng.suggest_name()
 
     @property
     @lru_cache(maxsize=None)
-    def email(self, allow_question=True):
+    def email(self):
         """Return the user's E-Mail Address."""
-        return self.userinfo['email']
+        try:
+            return self.userinfo['email']
+        except KeyError:
+            return None
 
     @property
     @lru_cache(maxsize=None)
-    def given_name(self, allow_question=True):
+    def given_name(self):
         """Return the user's given name. If none is provided, try to extract it from the full name."""
         return (self.userinfo.get('given_name')
                 or ' '.join(self.userinfo['name'].split(' ')[:-1]))
 
     @property
     @lru_cache(maxsize=None)
-    def family_name(self, allow_question=True):
+    def family_name(self):
         """Return the user's family name. If none is provided, try to extract it from the full name."""
         return (self.userinfo.get('family_name')
                 or self.userinfo.get('sn')
@@ -460,19 +547,19 @@ class UserInfo(Mapping):
 
     @property
     @lru_cache(maxsize=None)
-    def full_name(self, allow_question=True):
+    def full_name(self):
         """Return the user's full name. If none is provided, try to assemple it from the first and given name."""
         return (self.userinfo.get('name')
                 or ' '.join(filter(None, [self.given_name, self.family_name])))
 
     @property
     @lru_cache(maxsize=None)
-    def ssh_keys(self, allow_question=True):
+    def ssh_keys(self):
         """Return the user's SSH keys."""
         return self.credentials.get('ssh_key', [])
 
     @property
-    def entitlement(self, allow_question=True):
+    def entitlement(self):
         """Return the parsed entitlement attribute of the user. See `eduperson.Entitlement` for details."""
         attr = self.userinfo.get('eduperson_entitlement', [])
         if not isinstance(attr, list):
@@ -486,25 +573,43 @@ class UserInfo(Mapping):
 
         return filter(lambda x: x, map(try_entitlement, attr))
 
+    @property
+    def group(self):
+        """Return the unparsed group attribute of the user"""
+        attr = self.userinfo.get('groups', [])
+        if not isinstance(attr, list):
+            attr = [attr]
+        return attr
 
     @property
     @lru_cache(maxsize=None)
-    def groups(self, allow_question=True):
+    def groups(self):
         """Return the homogenised names of the groups the user should be a member of.
+        """
+        # A shitty way to see if the entitlement is empty or not:
+        if len([x for x in self.entitlement]) == 0:
+            logger.debug("Using plain groups from 'groups' claim")
+            grouplist = self.groups_from_grouplist()
+        else:
+            logger.debug("Using aarc-g002 groups from 'entitlements' claim")
+            grouplist = self.groups_from_entitlement()
+        return ([self._group_masked_for_bwidm(grp) for grp in grouplist])
 
-        These are extracted from the entitlement. Any additional 'group'-keys in the input are ignored.
+    def groups_from_entitlement(self):
+        """Gropus are extracted from the entitlement. Any additional 'group'-keys in the input are ignored.
 
         Group names are prefixed with the delegated namespace from the entitlement.
         """
-
         return set(filter(
             None,
-            ['{}_{}'.format(self._group_masked_for_bwidm(ns), self._group_masked_for_bwidm(grp))
-             for (ns, grp)
-             in chain.from_iterable(
+            ['{}_{}'.format(ns, grp) for (ns, grp) in chain.from_iterable(
                  (("-".join([ent.delegated_namespace] + ent.subnamespaces), grp) for grp in ent.all_groups)
                  for ent in self.entitlement)]
         ))
+    def groups_from_grouplist(self):
+        """Gropus are extracted from the groups claim
+        """
+        return (set( [grp for grp in self.group]))
 
     def _group_masked_for_bwidm(self, orig_grp):
         """Convert camelCase to snake_case, fixup beginning of name and replace invalid chars with a dash ('-')"""
@@ -533,19 +638,20 @@ class UserInfo(Mapping):
         grp = regex.sub('^9', 'nine_', grp)
 
         if grp != orig_grp:
-            logger.warning("Group name '{}' changed to '{}' for BWIDM compatibilty".format(orig_grp, grp))
+            if CONFIG.getboolean('messages', 'log_name_changes', fallback=True):
+                logger.warning("Group name '{}' changed to '{}' for BWIDM compatibilty".format(orig_grp, grp))
 
         return grp
 
     @property
     @lru_cache(maxsize=None)
-    def assurance(self, allow_question=True):
+    def assurance(self):
         """Return the assurance levels of the user."""
         return self.userinfo.get('eduperson_assurance', [])
 
     @property
     @lru_cache(maxsize=None)
-    def primary_group(self, allow_question=True):
+    def primary_group(self):
         config_group = CONFIG['ldf_adapter'].get("primary_group")
         if config_group:
             return config_group
@@ -554,13 +660,43 @@ class UserInfo(Mapping):
             for group in self.groups:
                 return group
         elif len(self.groups) > 1:
-            return self.value_or_ask(
-                self.userinfo.get(0), "primary_group",
-                "You are a member of multiple groups. Please select your desired primary group.",
-                allow_question, list(self.groups)
-            )
+            if self.allow_question:
+                return self.value_or_ask(
+                    self.userinfo.get(0), "primary_group",
+                    "You are a member of multiple groups. Please select your desired primary group.",
+                    self.allow_question, list(self.groups)
+                )
+            else: # make something up, regarding the primary group:
+                if CONFIG.getboolean('messages', 'log_primary_group_definition', fallback=True):
+                    logger.warning("/----- No primary group issue --------------------------------------------\\")
+                    logger.warning(F"    We have a user with mutiple primary groups, and no default primary group set in the config.")
+                    logger.warning(F"    Furthermore, we are in non-interactive mode, so we can't ask the user.")
+                    logger.warning(F"    Therefore, we just take the first group: '{list(self.groups)[0]}'")
+                    nl="\n                            "
+                    logger.warning(F"    Available groups are: {nl}{nl.join(self.groups)}")
+                    logger.warning("\\--------------------------------------------------------------------------------/")
+                return list(self.groups)[0]
+
         else:
-            raise Failure(message="No groups in userinfo and no global primary group configured")
+            old_answer = self.answers.get("primary_group", None)
+            if old_answer is not None:
+                return old_answer
+
+            else:  # still no group found.
+                fallback_group = CONFIG['ldf_adapter'].get("fallback_group", None)
+                if fallback_group:
+                    return fallback_group
+                else:
+                    logger.warning("Not a single group found; This may be ok, depending on the request type")
+            # raise Failure(message="No groups in userinfo and no global primary group configured")
+
+    @property
+    @lru_cache(maxsize=None)
+    def preferred_username(self):
+        """Return the prefrred username of the user."""
+        return self.userinfo.get('preferred_username', [])
+
+
 
     def value_or_ask(self, value, answer_name, question_text, allow_question, default=None):
         """Return the submitted answer, the default value or raise a questionaire."""
@@ -575,7 +711,7 @@ class UserInfo(Mapping):
 
 
     def __str__(self):
-        attrs = ("{} = {}".format(k, getattr(UserInfo, k).fget(self, allow_question=False)) for k in iter(self))
+        attrs = ("{} = {}".format(k, getattr(UserInfo, k).fget(self)) for k in iter(self))
 
         return "<UserInfo\n{}\n>".format("\n".join("\t{}".format(attr) for attr in attrs))
 
@@ -585,7 +721,7 @@ class UserInfo(Mapping):
     def __iter__(self):
         return (k for k in dir(UserInfo) if type(getattr(UserInfo, k)) is property)
 
-    def __len__(self, allow_questions=True):
+    def __len__(self):
         sum(1 for _ in filter(lambda k: type(getattr(UserInfo, k)) is property, dir(UserInfo)))
 
     def __hash__(self):
