@@ -8,6 +8,9 @@ import sys
 import json
 import logging
 
+# Must be before the first ldf_adapter import
+from feudal_globalconfig import globalconfig
+
 from ldf_adapter import User
 from ldf_adapter.results import ExceptionalResult
 
